/**
 * @file Fonctions pour interagir avec la BDD sigma et le LDAP.
 * @author akka vodol
 */
import knex from '../../db/knex_router';
import { renseignerSurUtilisateur, repliquerTOLdesIds, listerGroupes, listerMembres, listerAdministrateurs } from '../ldap/ldap_data';
import { exportAllDeclaration } from 'babel-types';

export { renseignerSurUtilisateur, repliquerTOLdesIds };

/**
 * @summary Renvoie tous les utilisateurs ayant des droits d'administrateur sur un groupe.
 * @function
 * @desc Les utilisateurs qui ont un droit d'administrateur sur un groupe sont ses administrateurs et les utilisateurs ayant droit d'admin sur son parent
 * @arg {String} uid - L'uid du groupe dont on veut les administrateurs. 
 * @return {Promise} Retour de requête knex. Promise qui renvera une liste de tous les utilisateurs ayant droit d'admin sur le groupe
 */
export const getUsersWithAdminRights = (user, groupUID) => {
    return listerAdministrateurs(user, groupUID).then(adminList => {
        if (typeof adminList == "undefined")
            return undefined;
        else
            return knex('groups').select('parentuid').where('uid', groupUID).then(req => {
                if (req[0].parentuid)
                    return getUsersWithAdminRights(user, req[0].parentuid).then(parentAdmins => {
                        return adminList.concat(parentAdmins);
                    });
                else
                    return adminList.concat(['anatole.romon']); // pour les besoins des tests, anatole romon a tout les droits
            });
    });
};

/**
 * @summary teste si un utilisateur a des droits
 * @desc Cette fonction effectue une requête knex. Elle gère l'arête de parenté.
 * @arg {Object} user - Objet contenant un attribut *uid* de type *string*. User représente l'utilisateur qui a effectué la requête. 
 * @return {Promise} Retour de requête knex. Liste de tous les groupes que l'utilisateur a le droit de voire.
 */
export const hasAdminRights = (user, groupUID) => {
    return getUsersWithAdminRights(user, groupUID).then(adminList => {
        return (typeof adminList != "undefined" && adminList.indexOf(user.uid) != -1);
    });
};

/**
 * @summary Renvoie tous les groupes visibles par l'utilisateur user
 * @desc Cette fonction effectue une requête knex. Elle gère l'arête de parenté.
 * @arg {Object} user - Objet contenant un attribut *uid* de type *string*. User représente l'utilisateur qui a effectué la requête. 
 * @return {Promise} Retour de requête knex. Liste de tous les groupes que l'utilisateur a le droit de voire.
 */
export const getAllVisibleGroups = (user) => {

    return listerGroupes(user, user.uid).then(group_ids => {
        if (typeof group_ids == "undefined")
            throw "invalid user";
        var membered_groups = qb => qb.select().from('groups').whereIn('uid', group_ids.concat(['kes']));
        /* membered_groups est un callback qui reçoit un argument qb représentant un objet de requête (un truc qui s'utilise comme knex())
        *  et effectue dessus la requête pour obtenir une table de tous les groupes dont user est membre
        *  with est ensuite utilisé pour donner à ce retour de requête l'alias 'membered_groups'.
        *  Cette table peut ainsi être JOIN avec la table groups. La condition de Join est donnée selon une syntaxe recopiée depuis la Doc Knex
        *  On récupère tous les groupes tels que le groupe est dans membered_groups ou le parent est dans membered_groups
        */
        return knex.with('membered_groups', membered_groups).distinct('groups.*').select().from('groups').innerJoin('membered_groups',
            function () {
                this.on('groups.uid', '=', 'membered_groups.uid').orOn('groups.parentuid', '=', 'membered_groups.uid');
            }
        );
    });
};


/**
 * @summary Teste si un utilisateur est membre d'un groupe
 * @arg {Object} user - Objet contenant un attribut *uid* de type *string*. User représente l'utilisateur qui a effectué la requête. 
 * @arg {Object} groupUID - L'id du groupe dont on veu savoir si l'utilisateur est membre. 
 * @return {Promise(Boolean)} Boolean indiquant si l'utilisateur est membre du groupe.
 */
export const isMember = (user, groupUID) => {
    return listerGroupes(user, user.uid).then(group_ids => group_ids && group_ids.indexOf(groupUID) != -1);
};

/**
 * @summary Renvoie un unique groupe, ssi ce groupe est visible par l'utilisateur
 * @desc Actuellement, la fonction effectue la même requête que *getAllVisibleGroups* et restreint au groupe demandé. Cette fonction peut être implémentée de manière plus efficace et plus chiante.
 * @arg {Object} user - Objet contenant un attribut *uid* de type *string*. User représente l'utilisateur qui a effectué la requête. 
 * @arg {String} uid - uid du groupe que l'on veut voire. 
 * @return {Promise(group)} Retour de requête knex. Le groupe demandé, si l'utilisateur a le droit de la voire.
 */
export const getGroupIfVisible = (user, groupUID) => {
    return listerGroupes(user, user.uid).then(group_ids => {
        if (typeof group_ids == "undefined")
            throw "invalid user";
        var membered_groups = qb => qb.select().from('groups').whereIn('uid', group_ids.concat(['kes']));
        return knex.with('membered_groups', membered_groups).distinct('groups.*').select().from('groups').innerJoin('membered_groups',
            function () {
                this.on('groups.uid', '=', 'membered_groups.uid').orOn('groups.parentuid', '=', 'membered_groups.uid');
            }
        ).where('groups.uid', groupUID).then(res => {
            return res[0];
        });
    });
};

/**
 * @summary Attribue un UID qui n'a pas encore été utilisé à un groupe
 * @desc RASifie le string initialUID si necessaire (ramené à de l'ASCCI sans espace), puis si l'uid est deja pris rajoute un n a la fin et reteste
 * @arg {String} uid - L'uid du groupe dont on veut les administrateurs. 
 * @return {Promise} Retour de requête knex. Promise qui renvera une liste de tous les utilisateurs ayant droit d'admin sur le groupe
 */
export const getAvailablegroupUID = (initialUID) => {
    let rasUID = initialUID.replace(' ', '_').replace(/\W/g, '').toLowerCase();
    return knex.from('groups').where('uid', rasUID).then(res => {
        if (res.length == 0) {
            return (rasUID);
        } else {
            return (getAvailablegroupUID(rasUID + 'n'));
        }
    });
};

/**
 * @summary Créé un groupe si les arguments sont tous valides
 * @desc Les arguments doivent être valides, sauf pour uid. Une clé uid valide sera générée dans tous les cas. 
 * Les authorisations de l'utilisateur ne sont pas vérifiées
 * On teste si l'utilisateur qui envoie la requête a des droits d'admin sur le parent du groupe qui doit être créé, avec la fonction *getUsersWithAdminRights*
 * Si un argument est invalide ou si l'utilisateur n'a pas les droits, la fonction renvoie une erreur
 * @arg {Object} user - L'utilisateur qui effectue la requête. 
 * @arg {Object} args - Les arguments envoyés à la mutation. Cf le schéma GraphQL 
 * @return {Promise} Retour de requête knex. Le groupe qui vient d'être créé. En cas d'echec, renvoie une erreur.
 */
export const createSubgroup = (user, args) => {
    if (typeof args.parentuid != 'string')
        throw "Illegal argument : parentuid must be a non null string";
    if (typeof args.name != 'string')
        throw "Illegal argument : name must be a non null string";

    return (getAvailablegroupUID(args.uid).then(rasUID => {
        // TODO : appeller une fonction de ldap_data pour y créer un groupe.
        return knex('groups').insert({
            uid: rasUID,
            parentuid: args.parentuid,
            createdAt: knex.fn.now(),
            updatedAt: this.createdAt,
            name: args.name,
            website: args.website,
            description: args.description,
            school: args.school
        }).then(res => {
            return getGroupIfVisible(user, rasUID);
        });
    }));
};

/**
 * @summary Créé un groupe si les arguments sont tous valides et l'utilisateur est authorisé
 * @desc Les arguments doivent être valides, sauf pour uid. Une clé uid valide sera générée dans tous les cas. 
 * On teste si l'utilisateur qui envoie la requête a des droits d'admin sur le parent du groupe qui doit être créé, avec la fonction *getUsersWithAdminRights*
 * Si un argument est invalide ou si l'utilisateur n'a pas les droits, la fonction renvoie une erreur
 * @arg {Object} user - L'utilisateur qui effectue la requête. 
 * @arg {Object} args - Les arguments envoyés à la mutation. Cf le schéma GraphQL 
 * @return {Promise} Retour de requête knex. Le groupe qui vient d'être créé. En cas d'echec, renvoie une erreur.
 */
export const createGroupIfLegal = (user, args) => {
    return hasAdminRights(user, args.parentuid).then(answer => {
        if (!answer)
            throw "illegal request : you must have admin rights over a group to create a subgroup of that group";
        return createSubgroup(user, args);
    });
};

export const getUserJoinGroupRequests = (user, recipientUID) => {
    return knex('user_join_group').select(/*'id', 'useruid', 'message'*/)/*.where('recipient', recipientUID)*/;
};

//Don't forget the argument user is the guy who makes the request, not the user we want
export const getUser = (user, userUID) => { 
    return new Promise( (resolve, reject) => {
        resolve({
            givenName : "Jean-Michel",
            lastName : "Samarchepas",
            uid : "jm",
            birthdate : "dans le turfu",
            mail : "email.email@email.com",
            phone : "06 60 06 60 06",
            groups : [],
            likes : [],
            address : "aussi dans le turfu"
        });
    });
    /*return renseignerSurUtilisateur(user, userUID).then(res => {
        return {
            givenName : res[0].givenName,
            lastName : res[0].lastName,
            uid : userUID,
            birthdate : res[0].brBirthdate,
            mail : res[0].mail,
            phone : res[0].telephoneNumber,
            groups : [],
            likes : [],
            address : res[0].brRoom
        };
    });*/
}